import getopt
import httplib
import os
import StringIO
import sys
import xml
from xml.dom import DOMImplementation, DocumentType
from xml.dom.ext import PrettyPrint
import xml.dom.minidom

class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

c_soapaddr = "127.0.0.1"
c_soapport = 80
c_soapcgi = "http://127.0.0.1/scripts/ttsoapcgi.exe"

m_comments = ""
m_files = ""

#Make a call to svnlook and use the 'changed' command to see which files
#were modified in the commit
def buildFileList(p_repo):

   #run the command and read in the output, errors
   #the command call will have to be changed in Linux (at least take out the '.exe',
   #might have to have entire path)
   stdin, stdout, stderr = os.popen3("svnlook.exe changed " + p_repo)
   output = stdout.read()
   desc = output
   errors = stderr.read()
   stdin.close(); stdout.close(); stderr.close()

   startpos = 0  
   endpos = desc.find("\n",0)

   filecount = 0
   #create a list to store all of the files
   fileList = []

   while endpos != -1:
      #The ouput given has the file name starting at the 4th position on each line
      fileChanged = desc[startpos+4:endpos]
      #Add the file to the list
      fileList[filecount:0] = [fileChanged]
      #reset the starting position
      startpos = endpos+1
      #reset the ending position
      endpos = desc.find("\n",startpos)
      #increment the filecount
      filecount = filecount+1

   return fileList

#Make a call to svnlook and use the 'log' command to retrieve the comments
#from the the commit so we can get the name of the project and defect list
 
def parseComments(p_repo):

   #run the command and read in the output, errors
   #the command call will have to be changed in Linux (at least take out the '.exe',
   #might have to have entire path)
   stdin, stdout, stderr = os.popen3("svnlook.exe log " + p_repo)
   output = stdout.read()
   desc = output
   errors = stderr.read()
   stdin.close(); stdout.close(); stderr.close()

   #First, make sure we find the token
   istart = desc.find("<ttp:")
   #if token found
   if istart >= 0:
      #find end of token
      iend = desc.find(">", istart)
      #extract from the comments the token
      ttpcomments = desc[istart+5:iend]
      icomma1=istart
      icomma2=0
      #this counter is to know when we read the first item of the token, which must be
      #the projectname
      myCounter=1

      projectName = ""
      defectnum = ""
      defectList = []
      defectCount = 0
      
      while icomma2 >= 0 :
         icomma2 = ttpcomments.find(",", icomma1)
         #icomma2 will be -1 once it does not find a comma in the remmainning portion,
         #so we know that is the last item in the token
         if icomma2 == -1:
            defectnum = ttpcomments[icomma1:len(ttpcomments)]
            defectList[defectCount:0] = [defectnum]
         else:
            if myCounter == 1:
               projectName = ttpcomments[icomma1:icomma2]
            else:
               defectnum = ttpcomments[icomma1:icomma2]
               defectList[defectCount:0] = [defectnum]
               defectCount = defectCount + 1
            #reset the starting position
            icomma1 = icomma2+1
            #increment the counter so we know it is not the project name
            myCounter=myCounter+1
            

   return projectName, defectList

def login(p_project, p_user, p_pwd):
   dom = DOMImplementation.getDOMImplementation ()
   typ = xml.dom.DocumentType.DocumentType ("Envelope", [], [], "", "")
   # this is not the namespace required by soap 1.2
   # but the service requires this namespace (from error message)
   doc = dom.createDocument ("http://schemas.xmlsoap.org/soap/envelope/", "Envelope", typ)

   env = doc.getElementsByTagName ("Envelope")[0]
   bdy = doc.createElement ("Body")
   env.appendChild (bdy)

   msg = doc.createElementNS ("urn:testtrack-interface", "DatabaseLogon")
   bdy.appendChild (msg)

   eleProject = doc.createElement("dbname")
   eleProject.setAttributeNode( doc.createAttribute("xsi:type"))
   eleProject.setAttribute("xsi:type", "xsd:string")
   msg.appendChild(eleProject)
   eleProject.appendChild(doc.createTextNode(str(p_project)))

   eleUser = doc.createElement("username")
   eleUser.setAttributeNode( doc.createAttribute("xsi:type"))
   eleUser.setAttribute("xsi:type", "xsd:string")
   msg.appendChild(eleUser)
   eleUser.appendChild(doc.createTextNode(str(p_user)))

   elePwd = doc.createElement("password")
   elePwd.setAttributeNode( doc.createAttribute("xsi:type"))
   elePwd.setAttribute("xsi:type", "xsd:string")
   msg.appendChild(elePwd)
   elePwd.appendChild(doc.createTextNode(str(p_pwd)))

   # Let's lock the defect for editing!
   buffer = StringIO.StringIO()
   PrettyPrint(doc, buffer)
   request = buffer.getvalue()
   #print request #DEBUG

   httpsvr = httplib.HTTPConnection(c_soapaddr, c_soapport)
   httpsvr.putrequest("POST", c_soapcgi)
   httpsvr.putheader("Content-Type", "text/xml")
   httpsvr.putheader("Content-Length", str(len(request)))
   httpsvr.endheaders()
   httpsvr.send(request)

   response = httpsvr.getresponse()
   session = response.read()
   print "DatabaseLogon", response.status, response.reason

   return session[session.find("<Cookie>")+8:session.find("</Cookie>")]

def logout(p_session):
   dom = DOMImplementation.getDOMImplementation ()
   typ = xml.dom.DocumentType.DocumentType ("Envelope", [], [], "", "")
   # this is not the namespace required by soap 1.2
   # but the service requires this namespace (from error message)
   doc = dom.createDocument ("http://schemas.xmlsoap.org/soap/envelope/", "Envelope", typ)

   env = doc.getElementsByTagName ("Envelope")[0]
   bdy = doc.createElement ("Body")
   env.appendChild (bdy)

   msg = doc.createElementNS ("urn:testtrack-interface", "DatabaseLogoff")
   bdy.appendChild (msg)

   eleCookie = doc.createElement("cookie")
   eleCookie.setAttributeNode( doc.createAttribute("xsi:type"))
   eleCookie.setAttribute("xsi:type", "xsd:int")
   msg.appendChild(eleCookie)
   eleCookie.appendChild(doc.createTextNode(str(p_session)))

   # Let's lock the defect for editing!
   buffer = StringIO.StringIO()
   PrettyPrint(doc, buffer)
   request = buffer.getvalue()

   httpsvr = httplib.HTTPConnection(c_soapaddr, c_soapport)
   httpsvr.putrequest("POST", c_soapcgi)
   httpsvr.putheader("Content-Type", "text/xml")
   httpsvr.putheader("Content-Length", str(len(request)))
   httpsvr.endheaders()
   httpsvr.send(request)

   response = httpsvr.getresponse()
   print "DatabaseLogoff", response.status, response.reason

def updateDefect(p_session, p_defectnumber, p_revnum, p_files):
   dom = DOMImplementation.getDOMImplementation ()
   typ = xml.dom.DocumentType.DocumentType ("Envelope", [], [], "", "")
   # this is not the namespace required by soap 1.2
   # but the service requires this namespace (from error message)
   doc = dom.createDocument ("http://schemas.xmlsoap.org/soap/envelope/", "Envelope", typ)

   env = doc.getElementsByTagName ("Envelope")[0]
   bdy = doc.createElement ("Body")
   env.appendChild (bdy)

   msg = doc.createElementNS ("urn:testtrack-interface", "editDefect")
   bdy.appendChild (msg)

   eleCookie = doc.createElement("cookie")
   eleCookie.setAttributeNode( doc.createAttribute("xsi:type"))
   eleCookie.setAttribute("xsi:type", "xsd:int")
   msg.appendChild(eleCookie)
   eleCookie.appendChild(doc.createTextNode(str(p_session)))

   eleNum = doc.createElement("defectNumber")
   eleNum.setAttributeNode( doc.createAttribute("xsi:type"))
   eleNum.setAttribute("xsi:type", "xsd:int")
   msg.appendChild(eleNum)
   eleNum.appendChild(doc.createTextNode(str(p_defectnumber)))

   eleSummary = doc.createElement("summary")
   eleSummary.setAttributeNode( doc.createAttribute("xsi:type"))
   eleSummary.setAttribute("xsi:type", "xsd:string")
   msg.appendChild(eleSummary)
   eleSummary.appendChild(doc.createTextNode(""))

   eleDnldAtt = doc.createElement("bDownloadAttachments")
   eleDnldAtt.setAttributeNode( doc.createAttribute("xsi:type"))
   eleDnldAtt.setAttribute("xsi:type", "xsd:boolean")
   msg.appendChild(eleDnldAtt)
   eleDnldAtt.appendChild(doc.createTextNode("false"))

   # Let's lock the defect for editing!
   buffer = StringIO.StringIO()
   PrettyPrint(doc, buffer)
   request = buffer.getvalue()

   httpsvr = httplib.HTTPConnection(c_soapaddr, c_soapport)
   httpsvr.putrequest("POST", c_soapcgi)
   httpsvr.putheader("Content-Type", "text/xml")
   httpsvr.putheader("Content-Length", str(len(request)))
   httpsvr.endheaders()
   httpsvr.send(request)

   response = httpsvr.getresponse()
   respDefect = response.read()

   # Replace 'editDefectRequest' with 'saveDefect' and we're ready to use this block to save changes.
   dom = xml.dom.minidom.parseString(respDefect.replace("editDefectResponse", "saveDefect"))

   eleSaveDefect = dom.getElementsByTagName("ttns:saveDefect")
   if (len(eleSaveDefect) == 0):
      return # defect wasn't found!

   eleCookie = dom.createElement("cookie")
   eleCookie.setAttributeNode( dom.createAttribute("xsi:type"))
   eleCookie.setAttribute("xsi:type", "xsd:int")
   eleCookie.appendChild(dom.createTextNode(str(p_session)))
   eleSaveDefect[0].appendChild(eleCookie)

   eleDefect = dom.getElementsByTagName("pDefect")[0]

   # first, go through and fix all of the item tags to have all of the required child elements for events.
   # The response doesn't include 'value' child elements for date fields w/o a value set.
   eleEvents = dom.getElementsByTagName("eventlist")
   if (len(eleEvents) > 0):
      eleEvents = eleEvents[0]
      records = eleEvents.getElementsByTagName("fieldlist") # there are multiple item elements and we only care about custom fields.
      iRecCount = len(records)
      for item in records:
         recItems = item.getElementsByTagName("item")
         for itemField in recItems:
            if (itemField.getAttribute("xsi:type") == "ttns:CDateField"): # only handle DateField types
               eleValue = itemField.getElementsByTagName("value")
               if (len(eleValue) == 0):
                  eleValue = dom.createElement("value")
                  itemField.appendChild( eleValue)
                  eleValue.appendChild( dom.createTextNode("0001-01-01")) # element has to have a closing bracket, or TTP can't handle it.
               else:
                  eleTextNode = eleValue[0].childNodes
                  if (len(eleTextNode) == 0):
                     eleValue[0].appendChild( dom.createTextNode("0001-01-01")) # element has to have a closing bracket, or TTP can't handle it.

   # second, go through and fix all of the item tags to have all of the required child elements for custom fields.
   # The response doesn't include 'value' child elements for date fields w/o a value set.
   eleCustom = dom.getElementsByTagName("customFieldList")
   if (len(eleCustom) > 0):
      eleCustom = eleCustom[0]
      records = eleCustom.getElementsByTagName("item")
      iRecCount = len(records)
      for item in records:
         if (item.getAttribute("xsi:type") == "ttns:CDateField"): # only handle DateField types
            eleValue = item.getElementsByTagName("value")
            if (len(eleValue) == 0):
               eleValue = dom.createElement("value")
               item.appendChild( eleValue)
               eleValue.appendChild( dom.createTextNode("0001-01-01")) # element has to have a closing bracket, or TTP can't handle it.
            else:
               eleTextNode = eleValue[0].childNodes
               if (len(eleTextNode) == 0):
                  eleValue[0].appendChild( dom.createTextNode("0001-01-01")) # element has to have a closing bracket, or TTP can't handle it.

   eleSCCI = dom.getElementsByTagName("pSCCFileList")
   # If the defect has no SCC files attached, then the response will not include the pSCCIFileList element so create it here.
   if (len(eleSCCI) == 0):
      eleSCCI = dom.createElement("pSCCFileList")
      eleDefect.appendChild(eleSCCI)
      eleSCCI.setAttributeNode( dom.createAttribute("xsi:type"))
      eleSCCI.setAttribute("xsi:type", "SOAP-ENC:Array")
   else:
      eleSCCI = eleSCCI[0]

   # Now, go through and fix all of item tags to have all of the required child elements. The
   # response doesn't include child elements of item if the value is blank ie. m-strFixedRevision
   # will not have an element for an item where it has no value. This won't work on the save, so fix it here.
   records = eleSCCI.getElementsByTagName("item")
   iRecCount = len(records)
   for item in records:
      eleFRev = item.getElementsByTagName("m-strFixedRevision")
      if (len(eleFRev) == 0):
         eleFRev = dom.createElement("m-strFixedRevision")
         item.appendChild( eleFRev)
         eleFRev.setAttributeNode( dom.createAttribute("xsi:type"))
         eleFRev.setAttribute("xsi:type", "xsd:string")
         eleFRev.appendChild( dom.createTextNode(""))

      eleFDate = item.getElementsByTagName("m-dateFixedTimestamp")
      if (len(eleFDate) == 0):
         eleFDate = dom.createElement("m-dateFixedTimestamp")
         item.appendChild( eleFDate)
         eleFDate.setAttributeNode( dom.createAttribute("xsi:type"))
         eleFDate.setAttribute("xsi:type", "xsd:dateTime")
         eleFDate.appendChild( dom.createTextNode(""))

   # Ok! The saveDefect request is now valid. All we have to do now is update it based on the current changelist.
   for file in p_files:
      records = eleSCCI.getElementsByTagName("item")
      fFound = 0
      if (len(records) > 0):
         for item in records:
            eleFName = item.getElementsByTagName("m-strFileName")
            if (len(eleFName) > 0):
               if (eleFName[0].firstChild.nodeValue == file):
                  # file is already attached, so just add to the rev and make sure all of the elements are present
                  fFound = 1

                  eleFRev = item.getElementsByTagName("m-strFixedRevision")
                  if (len(eleFRev) == 0):
                     eleFRev = dom.createElement("m-strFixedRevision")
                     item.appendChild( eleFRev)
                     eleFRev.setAttributeNode( dom.createAttribute("xsi:type"))
                     eleFRev.setAttribute("xsi:type", "xsd:string")
                     eleFRev.appendChild( dom.createTextNode(str(p_revnum)))
                  else:
                     if (len(eleFRev[0].firstChild.nodeValue) > 0):
                        eleFRev[0].firstChild.nodeValue = eleFRev[0].firstChild.nodeValue + ", " + str(p_revnum)
                     else:
                        eleFRev[0].firstChild.nodeValue = str(p_revnum)

      if fFound == 0: # File not already attached to defect, so let's add it here.
         iRecCount = iRecCount + 1
         eleItem = dom.createElement("item")
         eleSCCI.appendChild( eleItem)
         eleItem.setAttributeNode( dom.createAttribute("xsi:type"))
         eleItem.setAttribute("xsi:type", "ttns:CSCCFileRecord")

         eleRecId = dom.createElement("recordid")
         eleItem.appendChild( eleRecId)
         eleRecId.setAttributeNode( dom.createAttribute("xsi:type"))
         eleRecId.setAttribute("xsi:type", "xsd:int")
         eleRecId.appendChild( dom.createTextNode("0"))

         eleFName = dom.createElement("m-strFileName")
         eleItem.appendChild( eleFName)
         eleFName.setAttributeNode( dom.createAttribute("xsi:type"))
         eleFName.setAttribute("xsi:type", "xsd:string")
         eleFName.appendChild( dom.createTextNode(file))

         eleFRev = dom.createElement("m-strFixedRevision")
         eleItem.appendChild( eleFRev)
         eleFRev.setAttributeNode( dom.createAttribute("xsi:type"))
         eleFRev.setAttribute("xsi:type", "xsd:string")
         eleFRev.appendChild( dom.createTextNode(str(p_revnum)))

         eleFDate = dom.createElement("m-dateFixedTimestamp")
         eleItem.appendChild( eleFDate)
         eleFDate.setAttributeNode( dom.createAttribute("xsi:type"))
         eleFDate.setAttribute("xsi:type", "xsd:dateTime")
         eleFDate.appendChild( dom.createTextNode(""))

   # Update the pSCCIFileList element's type attribute to correctly reflect how many files are attached.
   eleSCCI.setAttributeNode( dom.createAttribute("SOAP-ENC:arrayType"))
   eleSCCI.setAttribute( "SOAP-ENC:arrayType", "ttns:CSCCFileRecord[" + str(iRecCount) + "]")

   # Let's save the defect!
   buffer = StringIO.StringIO ()
   PrettyPrint (dom, buffer)
   request = buffer.getvalue ()
   #print request #DEBUG

   httpsvr = httplib.HTTPConnection(c_soapaddr, c_soapport)
   httpsvr.putrequest("POST", c_soapcgi)
   httpsvr.putheader("Content-Type", "text/xml")
   httpsvr.putheader("Content-Length", str(len(request)))
   httpsvr.endheaders()
   httpsvr.send(request)

   resp = httpsvr.getresponse()
   print "saveDefect", resp.status, resp.reason

def main(argv=None):
   if argv is None:
      argv = sys.argv
   try:
      try:
         opts, args = getopt.getopt(argv[1:], "h", ["help"])
         p_project, m_defects = parseComments(args[0])
         p_files = buildFileList(args[0])
         log = file('D:\defectlist.log', 'w')
         for defect in m_defects:
            log.write(defect + "\n") 
            try:
               session = login(p_project, args[2], args[3])
               if session > 0:
                  updateDefect(session, defect, args[1], p_files)
                  logout(session)
            except getopt.error, msg:
               raise Usage(msg)      
      
      except getopt.error, msg:
         raise Usage(msg)
         # more code, unchanged
   except Usage, err:
      print >>sys.stderr, err.msg
      print >>sys.stderr, "for help use --help"
      return 2

if __name__ == "__main__":
   sys.exit(main())
